/*
**==============================================================================
**
** Open Management Infrastructure (OMI)
**
** Copyright (c) Microsoft Corporation
** 
** Licensed under the Apache License, Version 2.0 (the "License"); you may not 
** use this file except in compliance with the License. You may obtain a copy 
** of the License at 
**
**     http://www.apache.org/licenses/LICENSE-2.0 
**
** THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
** KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED 
** WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABLITY OR NON-INFRINGEMENT. 
**
** See the Apache 2 License for the specific language governing permissions 
** and limitations under the License.
**
**==============================================================================
*/

/* ifndef guards are ommitted on purpose (to allow including this again with different INTLSTR_DOMAIN) */

#include <pal/palcommon.h>

#if defined(_MSC_VER)
    #include <windows.h> /* needed for LocalFree, GetModuleHandle, etc. */
#endif

PAL_BEGIN_EXTERNC

/*
**==============================================================================
**
** Translating user-provided INTLSTR_DOMAIN into OS-specific _INTLSTR_DOMAIN
**
**==============================================================================
*/

#if defined (_INTLSTR_DOMAIN)
    #undef _INTLSTR_DOMAIN
#endif

#if defined(INTLSTR_DOMAIN)

    #if defined(_MSC_VER)

        #define _INTLSTR_DOMAIN (GetModuleHandleA(INTLSTR_DOMAIN))

    #else
    
        #define _INTLSTR_DOMAIN (INTLSTR_DOMAIN)
    
    #endif

#else

    #if defined(_MSC_VER)

        #define _INTLSTR_DOMAIN (GetModuleHandle(NULL))

    #else
    
        #define _INTLSTR_DOMAIN ("omi")
    
    #endif

#endif

/*
**==============================================================================
**
** Declarations of _Intlstr_GetString_LoadString() and _Intlstr_GetString_GetText() functions
**
** Intlstr_Define0 macro for translating user-provided localized string definitions into OS-specific
** Intlstr_<stringname> helper functions (written on top of _Intlstr_GetString_Xxx)
**
** Intlstr_FreeString inline function for releasing the string returned by functions above
**
**==============================================================================
*/

#if defined(_MSC_VER)

    _Ret_z_
    _Success_(return != NULL)
    PAL_Char* _Intlstr_GetString_LoadString(HINSTANCE hInstance, UINT uID);

    #define Intlstr_Define0(id, name, text) \
        _Ret_z_ \
        _Check_return_ \
        _Success_(return != NULL) \
        PAL_INLINE const PAL_Char* Intlstr_ ## name() \
        { \
            return _Intlstr_GetString_LoadString(_INTLSTR_DOMAIN, id); \
        } 

#elif defined(CONFIG_ENABLE_GETTEXT)

    _Ret_z_ \
    _Check_return_ \
    _Success_(return != NULL) \
    PAL_Char* _Intlstr_GetString_GetText(const char* domain_name, int id, const char* msgid);

    #define Intlstr_Define0(id, name, text) \
        _Ret_z_ \
        _Check_return_ \
        _Success_(return != NULL) \
        PAL_INLINE const PAL_Char* Intlstr_ ## name() \
        { \
            return _Intlstr_GetString_GetText(_INTLSTR_DOMAIN, id, text); \
        } 

#else

    #define Intlstr_Define0(id, name, text) \
        _Ret_z_ \
        _Check_return_ \
        _Success_(return != NULL) \
        PAL_INLINE const PAL_Char* Intlstr_ ## name() \
        { \
            return PAL_T(text); \
        } 

#endif


/*
**==============================================================================
**
** Declarations of _Intlstr_FormatString_FormatMessage() and _Intlstr_FormatString_gettext_and_snprintf() functions
**
** Intlstr_Define1/2/3 macros for translating user-provided localized string templates
** into OS-specific ntlstr_<stringname> helper functions (written on top of _Intlstr_FormatString_Xxx)
**
** Intlstr_FreeFormattedString inline function for releasing the string returned by functions above
** (yes, it is different from Intlstr_FreeString)
**
**==============================================================================
*/

#if defined(_MSC_VER)

    #include <windows.h> /* needed for LocalFree */

    _Ret_z_
    _Success_(return != NULL)
    PAL_Char* _Intlstr_FormatString_FormatMessage(HINSTANCE hInstance, UINT uID, ...);

    #define Intlstr_Define1(id, name, parameter1_type, parameter1_name, text) \
        _Ret_z_ \
        _Check_return_ \
        _Success_(return != NULL) \
        PAL_INLINE PAL_Char* Intlstr_ ## name(parameter1_type parameter1_name) \
        { \
            return _Intlstr_FormatString_FormatMessage(_INTLSTR_DOMAIN, id, parameter1_name); \
        } 

    #define Intlstr_Define2(id, name, parameter1_type, parameter1_name, parameter2_type, parameter2_name, text) \
        _Ret_z_ \
        _Check_return_ \
        _Success_(return != NULL) \
        PAL_INLINE PAL_Char* Intlstr_ ## name(parameter1_type parameter1_name, parameter2_type parameter2_name) \
        { \
            return _Intlstr_FormatString_FormatMessage(_INTLSTR_DOMAIN, id, parameter1_name, parameter2_name); \
        } 

    #define Intlstr_Define3(id, name, parameter1_type, parameter1_name, parameter2_type, parameter2_name, parameter3_type, parameter3_name, text) \
        _Ret_z_ \
        _Check_return_ \
        _Success_(return != NULL) \
        PAL_INLINE PAL_Char* Intlstr_ ## name(parameter1_type parameter1_name, parameter2_type parameter2_name, parameter3_type parameter3_name) \
        { \
            return _Intlstr_FormatString_FormatMessage(_INTLSTR_DOMAIN, id, parameter1_name, parameter2_name, parameter3_name); \
        } 

    PAL_INLINE void Intlstr_FreeFormattedString(_Frees_ptr_opt_ PAL_Char* localizedString)
    {
        LocalFree(localizedString);
    }

#elif defined(CONFIG_ENABLE_GETTEXT)

    _Ret_z_
    _Success_(return != NULL)
    PAL_Char* _Intlstr_FormatString_gettext_and_snprintf(const char* domain_name, int id, const char* msgid, ...);

    #define Intlstr_Define1(id, name, parameter1_type, parameter1_name, text) \
        _Ret_z_ \
        _Check_return_ \
        _Success_(return != NULL) \
        PAL_INLINE PAL_Char* Intlstr_ ## name(parameter1_type parameter1_name) \
        { \
            return _Intlstr_FormatString_gettext_and_snprintf(_INTLSTR_DOMAIN, id, text, parameter1_name); \
        }

    #define Intlstr_Define2(id, name, parameter1_type, parameter1_name, parameter2_type, parameter2_name, text) \
        _Ret_z_ \
        _Check_return_ \
        _Success_(return != NULL) \
        PAL_INLINE PAL_Char* Intlstr_ ## name(parameter1_type parameter1_name, parameter2_type parameter2_name) \
        { \
            return _Intlstr_FormatString_gettext_and_snprintf(_INTLSTR_DOMAIN, id, text, parameter1_name, parameter2_name); \
        }

    #define Intlstr_Define3(id, name, parameter1_type, parameter1_name, parameter2_type, parameter2_name, parameter3_type, parameter3_name, text) \
        _Ret_z_ \
        _Check_return_ \
        _Success_(return != NULL) \
        PAL_INLINE PAL_Char* Intlstr_ ## name(parameter1_type parameter1_name, parameter2_type parameter2_name, parameter3_type parameter3_name) \
        { \
            return _Intlstr_FormatString_gettext_and_snprintf(_INTLSTR_DOMAIN, id, text, parameter1_name, parameter2_name, parameter3_name); \
        }

    PAL_INLINE void Intlstr_FreeFormattedString(_Frees_ptr_opt_ PAL_Char* localizedString)
    {
        free(localizedString); 
    }

#else

    #include <pal/format.h>

    #define Intlstr_Define1(id, name, parameter1_type, parameter1_name, text) \
        _Ret_z_ \
        _Check_return_ \
        _Success_(return != NULL) \
        PAL_INLINE PAL_Char* Intlstr_ ## name(parameter1_type parameter1_name) \
        { \
            return Stprintf_StrDup(PAL_T(text), parameter1_name); \
        }

    #define Intlstr_Define2(id, name, parameter1_type, parameter1_name, parameter2_type, parameter2_name, text) \
        _Ret_z_ \
        _Check_return_ \
        _Success_(return != NULL) \
        PAL_INLINE PAL_Char* Intlstr_ ## name(parameter1_type parameter1_name, parameter2_type parameter2_name) \
        { \
            return Stprintf_StrDup(PAL_T(text), parameter1_name, parameter2_name); \
        }

    #define Intlstr_Define3(id, name, parameter1_type, parameter1_name, parameter2_type, parameter2_name, parameter3_type, parameter3_name, text) \
        _Ret_z_ \
        _Check_return_ \
        _Success_(return != NULL) \
        PAL_INLINE PAL_Char* Intlstr_ ## name(parameter1_type parameter1_name, parameter2_type parameter2_name, parameter3_type parameter3_name) \
        { \
            return Stprintf_StrDup(PAL_T(text), parameter1_name, parameter2_name, parameter3_name); \
        }

    PAL_INLINE void Intlstr_FreeFormattedString(_Frees_ptr_opt_ PAL_Char* localizedString)
    {
        free(localizedString); 
    }

#endif

PAL_END_EXTERNC


